@@ -3,32 +3,32 @@
 import datetime
 import os
 import unicodedata
-from itertools import groupby
+import uuid
 from io import BytesIO
+from itertools import groupby
 
 import genshi
 import genshi.template
+# XXX fix: https://genshi.edgewall.org/ticket/582
+from genshi.template.astutil import ASTCodeGenerator, ASTTransformer
 from lxml import etree
-from sql import Literal
+from sql import Literal, Null
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.pool import PoolMeta, Pool
-from trytond.model import (ModelSQL, ModelView, Workflow, fields, dualmethod,
-    Unique)
+from trytond.model import (
+    Index, ModelSQL, ModelView, Unique, Workflow, dualmethod, fields)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, If
-from trytond.transaction import Transaction
-from trytond.tools import reduce_ids, grouped_slice, sortable_values
-from trytond import backend
-
 from trytond.modules.account_payment.exceptions import ProcessError
 from trytond.modules.company import CompanyReport
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Eval, If
+from trytond.tools import (
+    grouped_slice, is_full_text, lstrip_wildcard, reduce_ids, sortable_values)
+from trytond.transaction import Transaction
 
 from .sepa_handler import CAMT054
 
-# XXX fix: https://genshi.edgewall.org/ticket/582
-from genshi.template.astutil import ASTCodeGenerator, ASTTransformer
 if not hasattr(ASTCodeGenerator, 'visit_NameConstant'):
     def visit_NameConstant(self, node):
         if node.value is None:
@@ -56,7 +56,7 @@ INITIATOR_IDS = [
     (None, ''),
     ('eu_at_02', "SEPA Creditor Identifier"),
     ('be_vat', "Belgian Enterprise Number"),
-    ('es_nif', "Spanish VAT Number"),
+    ('es_vat', "Spanish VAT Number"),
     ]
 
 
@@ -67,7 +67,7 @@ class Journal(metaclass=PoolMeta):
             context={
                 'company': Eval('company', -1),
                 },
-            depends=['company']),
+            depends={'company'}),
         'on_change_with_company_party')
     sepa_bank_account_number = fields.Many2One('bank.account.number',
         'Bank Account Number', states={
@@ -77,8 +77,7 @@ class Journal(metaclass=PoolMeta):
         domain=[
             ('type', '=', 'iban'),
             ('account.owners', '=', Eval('company_party')),
-            ],
-        depends=['process_method', 'company_party'])
+            ])
     sepa_payable_flavor = fields.Selection([
             (None, ''),
             ('pain.001.001.03', 'pain.001.001.03'),
@@ -88,8 +87,7 @@ class Journal(metaclass=PoolMeta):
             'required': Eval('process_method') == 'sepa',
             'invisible': Eval('process_method') != 'sepa',
             },
-        translate=False,
-        depends=['process_method'])
+        translate=False)
     sepa_receivable_flavor = fields.Selection([
             (None, ''),
             ('pain.008.001.02', 'pain.008.001.02'),
@@ -99,26 +97,22 @@ class Journal(metaclass=PoolMeta):
             'required': Eval('process_method') == 'sepa',
             'invisible': Eval('process_method') != 'sepa',
             },
-        translate=False,
-        depends=['process_method'])
+        translate=False)
     sepa_payable_initiator_id = fields.Selection(
         INITIATOR_IDS, "SEPA Payable Initiator Identifier",
         states={
             'invisible': Eval('process_method') != 'sepa',
             },
-        depends=['process_method'],
         help="The identifier used for the initiating party.")
     sepa_receivable_initiator_id = fields.Selection(
         INITIATOR_IDS, "SEPA Receivable Initiator Identifier",
         states={
             'invisible': Eval('process_method') != 'sepa',
             },
-        depends=['process_method'],
         help="The identifier used for the initiating party.")
     sepa_batch_booking = fields.Boolean('Batch Booking', states={
             'invisible': Eval('process_method') != 'sepa',
-            },
-        depends=['process_method'])
+            })
     sepa_charge_bearer = fields.Selection([
             ('DEBT', 'Debtor'),
             ('CRED', 'Creditor'),
@@ -127,8 +121,7 @@ class Journal(metaclass=PoolMeta):
             ], 'Charge Bearer', states={
             'required': Eval('process_method') == 'sepa',
             'invisible': Eval('process_method') != 'sepa',
-            },
-        depends=['process_method'])
+            })
 
     @classmethod
     def __setup__(cls):
@@ -151,6 +144,16 @@ class Journal(metaclass=PoolMeta):
                     values=['eu_at_02'],
                     where=column == 'sepa'))
 
+        # Migration from 6.8: es_nif renamed into es_vat
+        cursor.execute(*sql_table.update(
+                [sql_table.sepa_payable_initiator_id],
+                ['es_vat'],
+                where=sql_table.sepa_payable_initiator_id == 'es_nif'))
+        cursor.execute(*sql_table.update(
+                [sql_table.sepa_receivable_initiator_id],
+                ['es_vat'],
+                where=sql_table.sepa_receivable_initiator_id == 'es_nif'))
+
     @classmethod
     def default_company_party(cls):
         pool = Pool()
@@ -161,8 +164,7 @@ class Journal(metaclass=PoolMeta):
 
     @fields.depends('company')
     def on_change_with_company_party(self, name=None):
-        if self.company:
-            return self.company.party.id
+        return self.company.party if self.company else None
 
     @staticmethod
     def default_sepa_charge_bearer():
@@ -188,14 +190,20 @@ class Group(metaclass=PoolMeta):
         domain=[('company', '=', Eval('company', -1))],
         states={
             'invisible': ~Eval('sepa_messages'),
-            },
-        depends=['company'])
+            })
+    sepa_id = fields.Char("SEPA ID", readonly=True, size=35,
+        states={
+            'invisible': ~Eval('sepa_id'),
+            })
 
     @classmethod
     def __setup__(cls):
         super(Group, cls).__setup__()
         cls._buttons.update({
-                'generate_message': {},
+                'sepa_generate_message': {
+                    'invisible': Eval('process_method') != 'sepa',
+                    'depends': ['process_method'],
+                    },
                 })
 
     def get_sepa_template(self):
@@ -207,35 +215,60 @@ class Group(metaclass=PoolMeta):
     def process_sepa(self):
         pool = Pool()
         Payment = pool.get('account.payment')
+        Mandate = pool.get('account.payment.sepa.mandate')
         if self.kind == 'receivable':
-            mandates = Payment.get_sepa_mandates(self.payments)
-            for payment, mandate in zip(self.payments, mandates):
+            payments = list(self.payments)
+            mandates = Payment.get_sepa_mandates(payments)
+            Mandate.lock(list(filter(None, mandates)))
+            sequence_types = {}
+            for payment, mandate in zip(payments, mandates):
                 if not mandate:
                     raise ProcessError(
                         gettext('account_payment_sepa'
                             '.msg_payment_process_no_mandate',
                             payment=payment.rec_name))
-                # Write one by one because mandate.sequence_type must be
-                # recomputed each time
-                Payment.write([payment], {
-                        'sepa_mandate': mandate,
-                        'sepa_mandate_sequence_type': mandate.sequence_type,
-                        })
-        else:
-            for payment in self.payments:
-                if not payment.sepa_bank_account_number:
-                    raise ProcessError(
-                        gettext('account_payment_sepa'
-                            '.msg_payment_process_no_iban',
-                            payment=payment.rec_name))
-        self.generate_message(_save=False)
+                sequence_type = sequence_types.get(mandate)
+                if not sequence_type:
+                    sequence_type = mandate.sequence_type
+                    if sequence_type == 'FRST':
+                        sequence_types[mandate] = 'RCUR'
+                    else:
+                        sequence_types[mandate] = sequence_type
+                payment.sepa_mandate = mandate
+                payment.sepa_mandate_sequence_type = sequence_type
+            Payment.save(payments)
+        for payment in self.payments:
+            if not payment.sepa_bank_account_number:
+                raise ProcessError(
+                    gettext('account_payment_sepa'
+                        '.msg_payment_process_no_iban',
+                        payment=payment.rec_name))
+            if not payment.sepa_bank_account_number.account.bank:
+                bank_account = payment.sepa_bank_account_number.account
+                raise ProcessError(
+                    gettext('account_payment_sepa'
+                        '.msg_payment_process_no_bank',
+                        payment=payment.rec_name,
+                        bank_account=bank_account.rec_name))
+        to_write = []
+        for key, payments in self.sepa_payments:
+            to_write.append(payments)
+            to_write.append({
+                    'sepa_info_id': self.sepa_group_payment_id(key),
+                    })
+        if to_write:
+            Payment.write(*to_write)
+        self.sepa_id = uuid.uuid4().hex
+        self.sepa_generate_message(_save=False)
 
     @dualmethod
     @ModelView.button
-    def generate_message(cls, groups, _save=True):
+    def sepa_generate_message(cls, groups, _save=True):
         pool = Pool()
         Message = pool.get('account.payment.sepa.message')
         for group in groups:
+            if group.journal.process_method != 'sepa':
+                continue
             tmpl = group.get_sepa_template()
             if not tmpl:
                 raise NotImplementedError
@@ -255,17 +288,17 @@ class Group(metaclass=PoolMeta):
         return self.company.party
 
     def sepa_group_payment_key(self, payment):
-        key = (('date', payment.date),)
+        key = (
+            ('payment_info', payment.sepa_info_id),
+            ('date', payment.date),
+            )
         if self.kind == 'receivable':
             key += (('sequence_type', payment.sepa_mandate_sequence_type),)
             key += (('scheme', payment.sepa_mandate.scheme),)
         return key
 
     def sepa_group_payment_id(self, key):
-        payment_id = str(key['date'].toordinal())
-        if self.kind == 'receivable':
-            payment_id += '-' + key['sequence_type']
-        return payment_id
+        return key['payment_info'] or uuid.uuid4().hex
 
     @property
     def sepa_payments(self):
@@ -280,7 +313,23 @@ class Group(metaclass=PoolMeta):
 
     @property
     def sepa_message_id(self):
-        return self.number
+        return self.sepa_id or self.number
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
+        domain = super().search_rec_name(name, clause)
+        return [bool_op,
+            domain,
+            ('sepa_id', operator, code_value, *extra),
+            ]
 
 
 class Payment(metaclass=PoolMeta):
@@ -300,8 +349,7 @@ class Payment(metaclass=PoolMeta):
                 ('state', '=', 'validated'),
                 (),
                 )
-            ],
-        depends=['party', 'company', 'state', 'process_method', 'kind'])
+            ])
     sepa_mandate_sequence_type = fields.Char('Mandate Sequence Type',
         readonly=True)
     sepa_return_reason_code = fields.Char('Return Reason Code', readonly=True,
@@ -309,20 +357,22 @@ class Payment(metaclass=PoolMeta):
             'invisible': ((Eval('process_method') != 'sepa')
                 | (~Eval('sepa_return_reason_code')
                     & (Eval('state') != 'failed'))),
-            },
-        depends=['process_method', 'state'])
+            })
     sepa_return_reason_information = fields.Text('Return Reason Information',
         readonly=True,
         states={
             'invisible': ((Eval('process_method') != 'sepa')
                 | (~Eval('sepa_return_reason_information')
                     & (Eval('state') != 'failed'))),
-            },
-        depends=['process_method', 'state'])
+            })
     sepa_end_to_end_id = fields.Function(fields.Char('SEPA End To End ID'),
         'get_sepa_end_to_end_id', searcher='search_end_to_end_id')
     sepa_instruction_id = fields.Function(fields.Char('SEPA Instruction ID'),
         'get_sepa_instruction_id', searcher='search_sepa_instruction_id')
+    sepa_info_id = fields.Char("SEPA Info ID", readonly=True, size=35,
+        states={
+            'invisible': ~Eval('sepa_info_id'),
+            })
 
     @classmethod
     def copy(cls, payments, default=None):
@@ -391,10 +441,10 @@ class Payment(metaclass=PoolMeta):
             and self.sepa_return_reason_code
             and self.sepa_return_reason_information == '/RTYP/RJCT')
 
-    def create_clearing_move(self, date=None):
+    def _get_clearing_move(self, date=None):
         if not date:
             date = Transaction().context.get('date_value')
-        return super(Payment, self).create_clearing_move(date=date)
+        return super()._get_clearing_move(date=date)
 
     @classmethod
     def view_attributes(cls):
@@ -405,11 +455,28 @@ class Payment(metaclass=PoolMeta):
                     }),
             ]
 
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
+        domain = super().search_rec_name(name, clause)
+        return [bool_op,
+            domain,
+            ('sepa_info_id', operator, code_value, *extra),
+            ]
+
 
 class Mandate(Workflow, ModelSQL, ModelView):
     'SEPA Mandate'
     __name__ = 'account.payment.sepa.mandate'
-    party = fields.Many2One('party.party', 'Party', required=True, select=True,
+    party = fields.Many2One(
+        'party.party', "Party", required=True,
         states={
             'readonly': Eval('state').in_(
                 ['requested', 'validated', 'cancelled']),
@@ -417,7 +484,7 @@ class Mandate(Workflow, ModelSQL, ModelView):
         context={
             'company': Eval('company', -1),
             },
-        depends=['state', 'company'])
+        depends={'company'})
     account_number = fields.Many2One('bank.account.number', 'Account Number',
         ondelete='RESTRICT',
         states={
@@ -429,60 +496,55 @@ class Mandate(Workflow, ModelSQL, ModelView):
         domain=[
             ('type', '=', 'iban'),
             ('account.owners', '=', Eval('party')),
-            ],
-        depends=['state', 'party'])
+            ])
     identification = fields.Char('Identification', size=35,
         states={
             'readonly': Eval('identification_readonly', True),
             'required': Eval('state') == 'validated',
-            },
-        depends=['state', 'identification_readonly'])
+            })
     identification_readonly = fields.Function(fields.Boolean(
             'Identification Readonly'), 'get_identification_readonly')
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
+        'company.company', "Company", required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     type = fields.Selection([
             ('recurrent', 'Recurrent'),
             ('one-off', 'One-off'),
             ], 'Type',
         states={
             'readonly': Eval('state').in_(['validated', 'cancelled']),
-            },
-        depends=['state'])
+            })
     sequence_type_rcur = fields.Boolean(
         "Always use RCUR",
         states={
             'invisible': Eval('type') == 'one-off',
-            },
-        depends=['type'])
+            })
     scheme = fields.Selection([
             ('CORE', 'Core'),
             ('B2B', 'Business to Business'),
             ], 'Scheme', required=True,
         states={
             'readonly': Eval('state').in_(['validated', 'cancelled']),
-            },
-        depends=['state'])
+            })
     scheme_string = scheme.translated('scheme')
     signature_date = fields.Date('Signature Date',
         states={
             'readonly': Eval('state').in_(['validated', 'cancelled']),
             'required': Eval('state') == 'validated',
-            },
-        depends=['state'])
+            })
     state = fields.Selection([
             ('draft', 'Draft'),
             ('requested', 'Requested'),
             ('validated', 'Validated'),
             ('cancelled', 'Cancelled'),
-            ], 'State', readonly=True)
+            ], "State", readonly=True, sort=False)
     payments = fields.One2Many('account.payment', 'sepa_mandate', 'Payments')
     has_payments = fields.Function(fields.Boolean('Has Payments'),
         'get_has_payments')
+    is_first_payment = fields.Function(
+        fields.Boolean("Is First Payment"), 'get_is_first_payment')
 
     @classmethod
     def __setup__(cls):
@@ -518,6 +580,10 @@ class Mandate(Workflow, ModelSQL, ModelView):
             ('identification_unique', Unique(t, t.company, t.identification),
                 'account_payment_sepa.msg_mandate_unique_id'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                t, (t.state, Index.Equality()),
+                where=t.state.in_(['draft', 'requested'])))
 
     @classmethod
     def __register__(cls, module_name):
@@ -629,7 +695,7 @@ class Mandate(Workflow, ModelSQL, ModelView):
 
     @property
     def is_valid(self):
-        if self.state == 'validated':
+        if self.state == 'validated' and self.account_number.active:
             if self.type == 'one-off':
                 if not self.has_payments:
                     return True
@@ -641,9 +707,7 @@ class Mandate(Workflow, ModelSQL, ModelView):
     def sequence_type(self):
         if self.type == 'one-off':
             return 'OOFF'
-        elif not self.sequence_type_rcur and (not self.payments
-                or all(not p.sepa_mandate_sequence_type for p in self.payments)
-                or all(p.rejected for p in self.payments)):
+        elif not self.sequence_type_rcur and self.is_first_payment:
             return 'FRST'
         # TODO manage FNAL
         else:
@@ -666,6 +730,30 @@ class Mandate(Workflow, ModelSQL, ModelView):
 
         return has_payments
 
+    @classmethod
+    def get_is_first_payment(cls, mandates, name):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+        payment = Payment.__table__()
+        cursor = Transaction().connection.cursor()
+
+        is_first = dict.fromkeys([m.id for m in mandates], True)
+        for sub_ids in grouped_slice(mandates):
+            red_sql = reduce_ids(payment.sepa_mandate, sub_ids)
+            cursor.execute(*payment.select(
+                    payment.sepa_mandate, Literal(False),
+                    where=red_sql
+                    & (payment.sepa_mandate_sequence_type != Null)
+                    & ~(  # Same as property rejected
+                        (payment.state == 'failed')
+                        & ((payment.sepa_return_reason_code != Null)
+                            | (payment.sepa_return_reason_code != ''))
+                        & (payment.sepa_return_reason_information
+                            == '/RTYP/RJCT')),
+                    group_by=payment.sepa_mandate))
+            is_first.update(cursor)
+        return is_first
+
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
@@ -705,6 +793,12 @@ class Mandate(Workflow, ModelSQL, ModelView):
 class MandateReport(CompanyReport):
     __name__ = 'account.payment.sepa.mandate'
 
+    @classmethod
+    def get_context(cls, records, header, data):
+        context = super().get_context(records, header, data)
+        context['gettext'] = gettext
+        return context
+
 
 class Message(Workflow, ModelSQL, ModelView):
     'SEPA Message'
@@ -712,34 +806,36 @@ class Message(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Eval('state') != 'draft',
         }
-    _depends = ['state']
     message = fields.Binary('Message', filename='filename',
         file_id=file_id, store_prefix=store_prefix,
-        states=_states, depends=_depends)
+        states=_states)
     message_file_id = fields.Char("Message File ID", readonly=True)
     filename = fields.Function(fields.Char('Filename'), 'get_filename')
     type = fields.Selection([
             ('in', 'IN'),
             ('out', 'OUT'),
-            ], 'Type', required=True, states=_states, depends=_depends)
+            ], 'Type', required=True, states=_states)
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
+        'company.company', "Company", required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
-    origin = fields.Reference('Origin', selection='get_origin', select=True,
-        states=_states, depends=_depends)
+            })
+    origin = fields.Reference("Origin", selection='get_origin', states=_states)
     state = fields.Selection([
             ('draft', 'Draft'),
             ('waiting', 'Waiting'),
             ('done', 'Done'),
             ('cancelled', 'Cancelled'),
-            ], 'State', readonly=True, select=True)
+            ], "State", readonly=True, sort=False)
 
     @classmethod
     def __setup__(cls):
         super(Message, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t, (t.state, Index.Equality()),
+                where=t.state.in_(['draft', 'waiting'])))
         cls._transitions |= {
             ('draft', 'waiting'),
             ('waiting', 'done'),
@@ -768,32 +864,12 @@ class Message(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        pool = Pool()
-        Group = pool.get('account.payment.group')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
-        update = transaction.connection.cursor()
         table = cls.__table__()
 
         super(Message, cls).__register__(module_name)
 
-        # Migration from 3.2
-        if backend.TableHandler.table_exist(Group._table):
-            group_table = Group.__table_handler__(module_name)
-            if group_table.column_exist('sepa_message'):
-                group = Group.__table__()
-                cursor.execute(*group.select(
-                        group.id, group.sepa_message, group.company))
-                for group_id, message, company_id in cursor:
-                    update.execute(*table.insert(
-                            [table.message, table.type, table.company,
-                                table.origin, table.state],
-                            [[
-                                    message, 'out', company_id,
-                                    'account.payment.group,%s' % group_id,
-                                    'done']]))
-                group_table.drop_column('sepa_message')
-
         # Migration from 5.6: rename state canceled to cancelled
         cursor.execute(*table.update(
                 [table.state], ['cancelled'],
